#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import pandas as pd
import numpy as np
import pickle
from sklearn.utils import indexable
from sklearn.utils.validation import _num_samples

# User
from .exceptions import WrongInputException
#------------------------------------------------------------------------------
# Main
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Misc
#------------------------------------------------------------------------------
def break_links(x):
    """ Break link to x by copying it"""
    x = x.copy()
    return x

def isin(pattern, x, how="all", return_element_wise=True):
    """Check if any of the elements in 'pattern' is included in 'x'"""
    ALLOWED_HOW = ["all", "any"]
    
    # List iterable classes. Note that we cannot rely on __iter__ attribute, because then str will be iterable!
    ITERABLES = (list, tuple, np.ndarray, pd.Series)
    
    if how not in ALLOWED_HOW:
        raise WrongInputException(input_name="how",
                                  provided_input=how,
                                  allowed_inputs=ALLOWED_HOW)
        
    if not isinstance(x, ITERABLES):
        if how=="all":
            is_in = all(p == x for p in pattern)
        elif how=="any":
            is_in = any(p == x for p in pattern)
    else:
        is_in = [isin(pattern, x=y, how=how) for y in x]
        
        if not return_element_wise:
            is_in = any(is_in)
                
    return is_in

def remove_conditionally_invalid_keys(d, invalid_values=["deprecated"]):
    """ Remove keys from dictionary for which values contain specified values"""
    d = {k:v for k,v in d.items() if not isin(pattern=invalid_values,x=v,how="any",return_element_wise=False)}
        
    return d

def remove_invalid_attr(obj, invalid_attr):
    ALLOWED_INVALID_ATTR = [dict, list]
    
    if isinstance(invalid_attr, dict):
        for k,v in invalid_attr.items():
            if hasattr(obj, k):
                delattr(obj, k)
    elif isinstance(invalid_attr, list):
        for k in invalid_attr:
            if hasattr(obj, k):
                delattr(obj, k)
    else:
        raise WrongInputException(input_name="invalid_attr",
                                  provided_input=invalid_attr,
                                  allowed_inputs=ALLOWED_INVALID_ATTR)        

    return obj

def unlist_dict_values(d):
    
    # Turn all elements into list
    d = {key:(value if isinstance(value, list) else [value]) for key,value in d.items()}
    
    # Unlist if possible
    d = {key:(value if len(value)>1 else value[0]) for key,value in d.items()}
    
    return d
    
def get_unique_elements_from_list(l, keep_order=True):
    
    if keep_order:    
        seen = set()
        l_unique = [x for x in l if not (x in seen or seen.add(x))]
    else:
        l_unique = list(set(l))
        
    return l_unique



#------------------------------------------------------------------------------
# Save and load models
#------------------------------------------------------------------------------
# Save and load objects using pickle
def save_object_by_pickle(path,obj):
    with open(path, 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

def load_object_by_pickle(path):
    with open(path, 'rb') as f:
        return pickle.load(f)



class SingleSplit(object):
    """
    Split dataset in two fold (single split), similar to train-test split
    """
    # --------------------
    # Constructor function
    # --------------------
    def __init__(self,
                 test_size=0.25):
        self.test_size=test_size

    # --------------------
    # CLASS VARIABLES
    # --------------------        
    N_SPLITS = 1
        
    # --------------------
    # Public
    # --------------------        
    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        y : array-like of shape (n_samples,), default=None
            The target variable for supervised learning problems.
        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into
            train/test set.
        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """
        X, y, groups = indexable(X, y, groups)
        n_samples = _num_samples(X)    
        indices = np.arange(n_samples)

        n_test = int(self.test_size*n_samples)
        n_train = n_samples-n_test
        
        train = indices[0:n_train]
        test  = indices[n_train:n_samples]
        
        yield train, test

    def get_n_splits(self, X=None, y=None, groups=None):
        return self.N_SPLITS



#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
from sklearnex import patch_sklearn
patch_sklearn()
from abc import ABC, abstractmethod
import numpy as np
import pandas as pd
from sklearn.model_selection import ParameterSampler, GridSearchCV, KFold, TimeSeriesSplit
from copy import deepcopy

# User
from ..utils.tools import (SingleSplit,
                           remove_conditionally_invalid_keys,
                           unlist_dict_values, get_unique_elements_from_list)
from ..utils.model_params import get_param_grid_from_estimator, compose_model, update_params
from ..utils.sanity_check import check_param_grid, check_X_Y, check_X, check_estimator
from ..utils.exceptions import WrongInputException

#------------------------------------------------------------------------------
# Base ML Regressor
#------------------------------------------------------------------------------
class BaseMLRegressor(object):
    """
    This base class learns a conditional expectation of Y given X=x
    """
    # -------------------------------------------------------------------------
    # Constructor function
    # -------------------------------------------------------------------------    
    def __init__(self,
                 estimator,
                 param_grid=None,
                 cv_params={'scoring':None,
                            'n_jobs':None,
                            'refit':True,
                            'verbose':0,
                            'pre_dispatch':'2*n_jobs',
                            'error_score':np.nan,
                            'return_train_score':False},
                 fold_type="KFold",
                 n_cv_folds=5,
                 shuffle=False,
                 test_size=None,
                 max_n_models=50,
                 n_cf_folds=None,
                 verbose=False,
                 ):
        # Initialize inputs
        self.estimator = estimator
        self.param_grid = param_grid
        self.cv_params = cv_params
        self.n_cv_folds = n_cv_folds
        self.shuffle = shuffle
        self.test_size = test_size
        self.fold_type = fold_type
        self.max_n_models = max_n_models
        self.n_cf_folds = n_cf_folds
        self.verbose = verbose

        # ---------------------------------------------------------------------
        # Estimator
        # ---------------------------------------------------------------------
        # Check estimator type
        if isinstance(self.estimator, str):
            self.estimator = compose_model(estimator_name=self.estimator,
                                           perform_estimator_check=True,
                                           verbose=self.verbose)
        else:
            check_estimator(self.estimator)
        
        # ---------------------------------------------------------------------
        # Parameters
        # ---------------------------------------------------------------------
        self.param_grid = self._fix_params(estimator=self.estimator,
                                           param_grid=self.param_grid)

        # ---------------------------------------------------------------------
        # Misc
        # ---------------------------------------------------------------------
        # Define data splitter used in cross validation
        self.splitter = self._choose_splitter(n_folds=self.n_cv_folds,
                                              fold_type=self.fold_type,
                                              shuffle=self.shuffle,
                                              test_size=self.test_size)
            
        # Define cross-validated estimator
        self.estimator_cv = self._choose_estimator(estimator=self.estimator,
                                                   param_grid=self.param_grid,
                                                   cv_params=self.cv_params,                                                   
                                                   splitter=self.splitter,
                                                   max_n_models=self.max_n_models)
           

    # -------------------------------------------------------------------------
    # Class variables
    # -------------------------------------------------------------------------
    FOLD_TYPE_ALLOWED = ["KFold", "TimeSeriesSplit", "SingleSplit"]
    N_FOLDS_ALLOWED = [1, 2, "...", "N"]

    # -------------------------------------------------------------------------
    # Very Private functions
    # -------------------------------------------------------------------------
    def __getattr__(self, attrname):
        if not attrname in self.__dict__:
            return getattr(self.estimator_cv, attrname)

    # -------------------------------------------------------------------------
    # Private functions
    # -------------------------------------------------------------------------
    def _fix_params(self, estimator, param_grid):
        """ Fix parameters, including get parameters if not provided, remove invalids, convert to list, and remove duplicates"""
        
        # Obtain parameters if not provided
        if param_grid is None:
            param_grid = get_param_grid_from_estimator(estimator=estimator)
                        
        # Add default parameters if for some reason not specified    
        param_grid = update_params(old_param=estimator.get_params(),
                                        new_param=param_grid)
        
        # Remove invalid keys
        param_grid = remove_conditionally_invalid_keys(d=param_grid,
                                                        invalid_values=["deprecated"])
                
        # Set param_grid values to list if not already list
        param_grid = {k: v if isinstance(v, list) else v.tolist() if isinstance(v, np.ndarray) else [v] for k, v in param_grid.items()}

        # Remove duplicates
        param_grid = {k: get_unique_elements_from_list(l=v,keep_order=True) for k,v in param_grid.items()}
        
        # Check parameter grid
        check_param_grid(param_grid)
        
        return param_grid
        
    def _choose_splitter(self, n_folds=2, fold_type="KFold", shuffle=True, test_size=0.25):
        """ Define the split function that splits the data for cross-validation"""
        if (n_folds==1) or (fold_type=="SingleSplit"):
            if test_size is None:
                raise Exception("""
                                When 'n_folds==1', we automatically instantiate a SingleSplit.
                                This requires 'test_size' to be a float and not 'None'
                                """)                
            else:
                splitter = SingleSplit(test_size=test_size)
            
        elif n_folds>=2:
            if fold_type=="KFold":
                splitter = KFold(n_splits=n_folds, random_state=None, shuffle=shuffle)
            elif fold_type=="TimeSeriesSplit":
                splitter = TimeSeriesSplit(n_splits=n_folds, max_train_size=None, test_size=None, gap=0)
            else:
                raise WrongInputException(input_name="fold_type",
                                          provided_input=fold_type,
                                          allowed_inputs=self.FOLD_TYPE_ALLOWED)
        else:
            raise WrongInputException(input_name="n_folds",
                                      provided_input=n_folds,
                                      allowed_inputs=self.N_FOLDS_ALLOWED)        
        
        return splitter
        
    def _choose_estimator(self, estimator, param_grid, cv_params, splitter, max_n_models):
        """
        Choose between grid search or randomized search, or simply the estimator if only one parametrization is provided.
        Note that if the estimator ends with "CV", we assume it has a native implementation as in scikit-learn. We use that in this case.
        """       
        if type(estimator).__name__[-2:]=="CV":
            # Unlist if possible
            param_grid = unlist_dict_values(d=param_grid)
            
            # Get first element of parameter grid
            # param_grid = {k: param_grid.get(k,None)[0] for k in param_grid.keys()}
            # Set estimator
            estimator_cv = estimator
                        
            # Update params
            param_grid = update_params(old_param=param_grid,
                                       new_param={
                                           "cv":splitter
                                           }
                                       )
            
            # Set parameters
            estimator_cv.set_params(**param_grid)
            
        else:
            # Compute number of models
            n_models = np.prod(np.array([len(v) for k,v in param_grid.items()]))
                        
            if n_models>1:            
                
                if n_models>max_n_models:
                    
                    # max_n_models=3
                    # param_grid={"a":[0,"aaa",2,3],
                    #             "b":["a0a","bbb",5,6]}
                                        
                    # Select (max_n_models-1) combinations of parameters AND add default (being the first parameter)
                    self.param_sampled = list(ParameterSampler(param_distributions=param_grid,
                                                          n_iter=max_n_models-1))

                    self.param_default = {k: param_grid.get(k,None)[0] for k in param_grid.keys()}                    
                    
                    # Combine lists of dicts
                    param_grid = [self.param_default]+self.param_sampled
                    
                    # Listify all values in dicts
                    param_grid = [{k:[v] for k,v in l.items()} for l in param_grid]
                    
                    # Convert to dict of lists
                    # param_grid = pd.DataFrame(param_grid).to_dict('list')
                    
                # Grid search over all parameters provided (sampled or not)    
                estimator_cv = GridSearchCV(estimator=estimator,
                                            param_grid=param_grid,
                                            cv=splitter,
                                            **cv_params)
            
            else:
                # If param_grid leads to one single model (n_models==1), there's no need to set of cross validation. In this case, just initialize the model and set parameters
                estimator_cv = estimator
                param_grid = {k: param_grid.get(k,None)[0] for k in param_grid.keys()}
                
                # Set parameters
                estimator_cv.set_params(**param_grid)        
            
        return estimator_cv
        
    # -------------------------------------------------------------------------
    # Public functions
    # -------------------------------------------------------------------------
    def fit(self,X,y):
        
        # Check X and Y
        X, y = check_X_Y(X, y)
                        
        # Estimate f in Y0 = f(X) + eps
        self.estimator_cv.fit(X=X,y=y)
        
        # Mean cross-validated score of the best_estimator
        if hasattr(self.estimator_cv, "best_score_"):     
            self.best_mean_cv_score_ = self.estimator_cv.best_score_
        else:
            self.best_mean_cv_score_ = None
                
        return self

    def get_params(self,deep=True):            
        return self.estimator_cv.get_params(deep=deep)

    def predict(self,X):
        
        # Check X and Y
        X = check_X(X)
                        
        # Estimate f in Y0 = f(X) + eps
        y_hat = self.estimator_cv.predict(X=X)
                        
        return y_hat
        
    def score(self,X,y):
        
        # Check X and Y
        X, y = check_X_Y(X, y)
                        
        score = self.estimator_cv.score(X=X,y=y)
                
        return score

    def set_params(self,**params):
        self.estimator_cv.set_params(**params)

    def cross_fit(self,X,y,shuffle=False):
        
        # Check X and Y
        X, y = check_X_Y(X, y)

        # Check n_cf_folds
        if self.n_cf_folds is None:
            raise Exception("""
                            When calling 'cross_fit()', one has to intentionally provide 'n_cf_folds' in the __init__
                            That is, 'n_cf_folds' is required to be an integer and not 'None'
                            """)     

        # Split into folds
        k_folds = KFold(n_splits=self.n_cf_folds, shuffle=shuffle)
        
        # Pre-allocate indices for the TEST set
        self.indices_ = np.zeros_like(a=y)
        self.y_pred_cf_ = np.zeros_like(a=y)
        
        for i,(train_index, test_index) in enumerate(k_folds.split(X)):
            
            # Increase cnt
            i += 1
            
            print(f"Cross-fitting fold {i}/{self.n_cf_folds}")
            
            # Fill indices
            self.indices_[test_index] = i

            # Copy estimator
            estimator_cf = deepcopy(self.estimator_cv)

            # Fit model using train indices
            estimator_cf.fit(X=X[train_index],
                             y=y[train_index])
        
            # Predict using test indices
            self.y_pred_cf_[test_index] = estimator_cf.predict(X=X[test_index])
    
            # Store estimator as attr
            setattr(self,f"estimator_{i}",estimator_cf)
    
        # Compute residuals
        self.y_res_cf_ = y - self.y_pred_cf_
    
        return self
    
    
        


#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import numpy as np
import xgboost as xgb
import lightgbm as lgbm

#------------------------------------------------------------------------------
# XGBoost
#------------------------------------------------------------------------------
class XGBRegressor(xgb.XGBRegressor):
    """
    This class copies verbatim the XGBoost regressor
    See: https://xgboost.readthedocs.io/en/latest/python/python_api.html#module-xgboost.sklearn
    """
    # -------------------------------------------------------------------------
    # Constructor function
    # -------------------------------------------------------------------------
    def __init__(self,
                  n_estimators=200,                                             # Default 100 
                  max_depth=None,
                  learning_rate=1,
                  verbosity=0,
                  objective='reg:squarederror',
                  booster=None,
                  tree_method=None,
                  n_jobs=1,
                  gamma=None,
                  min_child_weight=None,
                  max_delta_step=None,
                  subsample=0.8,
                  colsample_bytree=None,
                  colsample_bylevel=None,
                  colsample_bynode=0.8,
                  reg_alpha=None,
                  reg_lambda=1e-05,
                  scale_pos_weight=None,
                  base_score=None,
                  random_state=1991,
                  missing=np.nan,
                  num_parallel_tree=None,
                  monotone_constraints=None,
                  interaction_constraints=None,
                  importance_type='gain',
                  gpu_id=None,
                  validate_parameters=None,
                  ):
        # self.n_estimators=n_estimators                           
        # self.max_depth=max_depth
        # self.learning_rate=learning_rate
        # self.verbosity=verbosity
        # self.booster=booster
        # self.tree_method=tree_method
        # self.n_jobs=n_jobs
        # self.gamma=gamma
        # self.min_child_weight=min_child_weight
        # self.max_delta_step=max_delta_step
        # self.subsample=subsample
        # self.colsample_bytree=colsample_bytree
        # self.colsample_bylevel=colsample_bylevel
        # self.colsample_bynode=colsample_bynode
        # self.reg_alpha=reg_alpha
        # self.reg_lambda=reg_lambda
        # self.scale_pos_weight=scale_pos_weight
        # self.base_score=base_score
        # self.random_state=random_state
        # self.missing=missing
        # self.num_parallel_tree=num_parallel_tree
        # self.monotone_constraints=monotone_constraints
        # self.interaction_constraints=interaction_constraints
        # self.importance_type=importance_type
        # self.gpu_id=gpu_id
        # self.validate_parameters=validate_parameters
        super().__init__(
            n_estimators=n_estimators,                                             
            max_depth=max_depth,
            learning_rate=learning_rate,
            verbosity=verbosity,
            booster=booster,
            tree_method=tree_method,
            n_jobs=n_jobs,
            gamma=gamma,
            min_child_weight=min_child_weight,
            max_delta_step=max_delta_step,
            subsample=subsample,
            colsample_bytree=colsample_bytree,
            colsample_bylevel=colsample_bylevel,
            colsample_bynode=colsample_bynode,
            reg_alpha=reg_alpha,
            reg_lambda=reg_lambda,
            scale_pos_weight=scale_pos_weight,
            base_score=base_score,
            random_state=random_state,
            missing=missing,
            num_parallel_tree=num_parallel_tree,
            monotone_constraints=monotone_constraints,
            interaction_constraints=interaction_constraints,
            importance_type=importance_type,
            gpu_id=gpu_id,
            validate_parameters=validate_parameters
            )

# # Lazy implementation:
# class XGBRegressor(xgb.XGBRegressor):    
#     def __init__(self, **kwargs):  
#         super().__init__(**kwargs)


#------------------------------------------------------------------------------
# LightGBM
#------------------------------------------------------------------------------
class LGBMegressor(lgbm.LGBMRegressor):
    """
    This class copies verbatim the LightGBM regressor
    See: https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html#lightgbm-lgbmregressor
    """
    # -------------------------------------------------------------------------
    # Constructor function
    # -------------------------------------------------------------------------
    def __init__(self,
                 boosting_type='gbdt',
                 num_leaves=31,
                 max_depth=-1,
                 learning_rate=0.1,
                 n_estimators=100,
                 subsample_for_bin=200000,
                 objective='regression',
                 class_weight=None,
                 min_split_gain=0.0,
                 min_child_weight=0.001,
                 min_child_samples=20,
                 subsample=1.0,
                 subsample_freq=0,
                 colsample_bytree=1.0,
                 reg_alpha=0.0,
                 reg_lambda=0.0,
                 random_state=None,
                 n_jobs=1,
                 silent='warn',
                 importance_type='split'
                 ):
        super().__init__(
            boosting_type=boosting_type,
            num_leaves=num_leaves,
            max_depth=max_depth,
            learning_rate=learning_rate,
            n_estimators=n_estimators,
            subsample_for_bin=subsample_for_bin,
            objective=objective,
            class_weight=class_weight,
            min_split_gain=min_split_gain,
            min_child_weight=min_child_weight,
            min_child_samples=min_child_samples,
            subsample=subsample,
            subsample_freq=subsample_freq,
            colsample_bytree=colsample_bytree,
            reg_alpha=reg_alpha,
            reg_lambda=reg_lambda,
            random_state=random_state,
            n_jobs=n_jobs,
            silent=silent,
            importance_type=importance_type
            )


#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import numpy as np
from sklearn import (dummy,ensemble,gaussian_process,isotonic,kernel_ridge,
                     linear_model,neighbors,neural_network,svm,tree,)
from itertools import compress

# User
from ..estimator import (boosting)
from .sanity_check import check_estimator
from .exceptions import WrongInputException

#------------------------------------------------------------------------------
# Model tools
#------------------------------------------------------------------------------
def compose_model(estimator_name, perform_estimator_check=True, verbose=True):
        
        # Modules from scikit-learn used for regression
        SCIKIT_REGRESSION_MODULES = ["dummy","ensemble","gaussian_process",
                                     "isotonic","kernel_ridge","linear_model",
                                     "neighbors","neural_network","svm","tree",]
        
        # Internal modules used for regression
        NONSCIKIT_REGRESSION_MODULES = ["boosting",]
        
        REGRESSION_MODULES = [SCIKIT_REGRESSION_MODULES,NONSCIKIT_REGRESSION_MODULES]
        
        # Count the number of modules in which we find the model name
        n_modules_found_in = 0
        
        for reg_module in REGRESSION_MODULES:
            
            # Check if 'estimator_name' is an attribute of any module
            module_check = [hasattr(eval(lib), estimator_name) for lib in reg_module]
        
            is_found = any(module_check)
    
            # Compose model            
            if is_found:
                
                # Check counter
                if n_modules_found_in>1:
                    raise Exception(f"""Algorithm '{estimator_name}' is part of multiple modules, namely at least two of
                                    \n{SCIKIT_REGRESSION_MODULES}
                                    \n{NONSCIKIT_REGRESSION_MODULES}
                                    """)
                
                # Locate module
                module = list(compress(reg_module, module_check))[0]

                # Instantiate model
                estimator = eval(module+"."+estimator_name+"()")

                # Increase model counter
                n_modules_found_in += 1

        if n_modules_found_in==0:
            raise Exception(f"""Algorithm '{estimator_name}' is NOT part any module, namely neither of
                            \n{SCIKIT_REGRESSION_MODULES}
                            \n{NONSCIKIT_REGRESSION_MODULES}
                            """)

        # Model checks
        if perform_estimator_check:
            check_estimator(estimator=estimator)

        return estimator

#------------------------------------------------------------------------------
# Parameter tools
#------------------------------------------------------------------------------
def update_params(old_param, new_param, errors="raise"):
    """ Update 'old_param' with 'new_param'
    """
    # Copy old param
    updated_param = old_param.copy()
    
    for k,v in new_param.items():
        if k in old_param:
            updated_param[k] = v
        else:
            if errors=="raise":
                raise Exception(f"Parameters {k} not recognized as a default parameter for this estimator")
            else:
                pass
            
    return updated_param

def isinstance_double(obj, class_or_tuple, how="all"):
    ALLOWED_HOW = ["all", "any"]
    
    if how=="all":
        is_instance = isinstance(obj,class_or_tuple) and (obj.__class__.__name__==class_or_tuple.__name__)
    elif how=="any":
        is_instance = isinstance(obj,class_or_tuple) or (obj.__class__.__name__==class_or_tuple.__name__)    
    else:
        raise WrongInputException(input_name="how",
                                  provided_input=how,
                                  allowed_inputs=ALLOWED_HOW)
    
    return is_instance


def get_param_grid_from_estimator(estimator):
    """
    Get the corresponding parameter grid from the estimator.
    The default parameter is *ALWAYS* the FIRST parameter listed, *UNLESS* the parameter is followed by a comment, e.g., "Default 100"
    Read the scikit-learn reference here: https://scikit-learn.org/stable/modules/classes.html# 
    """
    #--------------------------------------------------------------------------
    # from ..estimator.boosting
    #--------------------------------------------------------------------------        
    if isinstance_double(estimator, boosting.XGBRegressor):
        param_grid = {
            "n_estimators":[100,200,500],
            "max_depth":[None,2,4,8,16],
            "learning_rate":[1, 0.8, 0.5, 0.3, 0.1, 0.01],
            "verbosity":0,
            # "objective":'reg:squarederror', # Leads to RuntimeError: Cannot clone object XGBRegressor(...), as the constructor either does not set or modifies parameter objective
            "booster":None,
            "tree_method":None,
            "gamma":[None, 0, 0.5],
            "min_child_weight":[None,1,2,4,8],
            "max_delta_step":None,
            "subsample":[0.8,1,0.5],
            "colsample_bytree":[None,1,2/3,1/3],
            "colsample_bylevel":None,
            "colsample_bynode":[0.8,1,0.5],
            "reg_alpha":None,
            "reg_lambda":[1e-05,0],
            "scale_pos_weight":None,
            "base_score":None,
            "num_parallel_tree":None,
            "monotone_constraints":None,
            "interaction_constraints":None,
            "importance_type":'gain',
            }
    
    elif isinstance_double(estimator, boosting.LGBMegressor):
        param_grid = {
            "boosting_type":'gbdt',
            "num_leaves":[31,11,21,41],
            "max_depth":[-1,2,4,8,16],
            "learning_rate":[0.1,0.8, 0.5, 0.3, 0.01],
            "n_estimators":[100,200,500],
            "subsample_for_bin":200000,
            "objective":'regression',
            "class_weight":None,
            "min_split_gain":0.0,
            "min_child_weight":0.001,
            "min_child_samples":[20,4,8,16],
            "subsample":[1.0,0.8,0.5],
            "subsample_freq":0,
            "colsample_bytree":[1.0,2/3,1/3],
            "reg_alpha":[0.0,1e-05],
            "reg_lambda":[0.0,1e-05],
            "importance_type":'split'
            }  
    
    #--------------------------------------------------------------------------
    # from sklearn.dummy
    #--------------------------------------------------------------------------
    elif isinstance_double(estimator, dummy.DummyRegressor):
        param_grid = {
            "strategy":["mean","median"],
            }

    #--------------------------------------------------------------------------
    # from sklearn.ensemble
    #--------------------------------------------------------------------------
    # elif isinstance_double(estimator, ensemble.AdaBoostRegressor):
    #     raise NotImplementedError()
        
    # elif isinstance_double(estimator, ensemble.BaggingRegressor):
    #     raise NotImplementedError()
        
    elif isinstance_double(estimator, ensemble.ExtraTreesRegressor):
        param_grid = {
            "n_estimators":200,                                                # Default 100
            'criterion':'squared_error',
            'max_depth':[None,2,4,8,16],
            'min_samples_split':[2,4,8,16],
            'min_samples_leaf':[1,2,4,8],
            'min_weight_fraction_leaf':0.0,
            'max_features':['sqrt',1/4,1/3,1/2,2/3,'log2','auto'],             # Default 'auto'
            'max_leaf_nodes':None,
            'min_impurity_decrease':0.0,
            'bootstrap':False,
            'oob_score':False,
            'ccp_alpha':0.0,
            'max_samples':None
            }   

    elif isinstance_double(estimator, ensemble.GradientBoostingRegressor):
        param_grid = {
            "loss":'squared_error',
            'learning_rate':[0.1,0.8, 0.5, 0.3, 0.01],
            'n_estimators':[100,200,500],
            'subsample':[1.0,0.8,0.5],
            'criterion':'friedman_mse',
            'min_samples_split':[2,4,8,16],
            'min_samples_leaf':[1,2,4,8],
            'min_weight_fraction_leaf':0.0,
            'max_depth':[3,2,4,8,16],
            'min_impurity_decrease':0.0,
            'max_features':['sqrt',1/4,1/3,1/2,2/3,'log2',None],               # Default None
            'alpha':0.9,
            'verbose':0,
            'max_leaf_nodes':None,
            'validation_fraction':0.1,
            'n_iter_no_change':None,
            'tol':0.0001,
            'ccp_alpha':0.0
            }
        
    elif isinstance_double(estimator, ensemble.RandomForestRegressor):
        param_grid = {
            "n_estimators":200,                                                # Default 100
            "criterion":'squared_error',
            "max_depth":[None,2,4,8,16],
            "min_samples_split":[2,4,8,16],
            "min_samples_leaf":[1,2,4,8],
            "min_weight_fraction_leaf":0.0,
            "max_features":['sqrt',1/4,1/3,1/2,2/3,'log2','auto'],             # Default 'auto'    
            "max_leaf_nodes":None,
            "min_impurity_decrease":0.0,
            "bootstrap":True,
            "oob_score":False,
            "n_jobs":None,
            "random_state":None,
            "verbose":0,
            "warm_start":False,
            # "ccp_alpha":0.0, # Leads to RuntimeError: Cannot clone object RandomForestRegressor(...), as the constructor either does not set or modifies parameter ccp_alpha
            "max_samples":None
            }

    # elif isinstance_double(estimator, ensemble.HistGradientBoostingRegressor):
    #     raise NotImplementedError()    

    #--------------------------------------------------------------------------
    # from sklearn.gaussian_process
    #--------------------------------------------------------------------------
    # elif isinstance_double(estimator, gaussian_process.GaussianProcessRegressor):
    #     raise NotImplementedError()    

    #--------------------------------------------------------------------------
    # from sklearn.isotonic
    #--------------------------------------------------------------------------
    # elif isinstance_double(estimator, isotonic.IsotonicRegression):
    #     raise NotImplementedError()    
    
    #--------------------------------------------------------------------------
    # from sklearn.kernel_ridge
    #--------------------------------------------------------------------------
    # elif isinstance_double(estimator, kernel_ridge.KernelRidge):
    #     raise NotImplementedError()    
    
    #--------------------------------------------------------------------------
    # from sklearn.linear_model
    #--------------------------------------------------------------------------
    elif isinstance_double(estimator, linear_model.LinearRegression):
        param_grid = {
            "fit_intercept":True,
            "copy_X":True,
            "n_jobs":None,
            "positive":False
            }
                
    elif isinstance_double(estimator, linear_model.Ridge):
        param_grid = {
            'alpha':[1.0,0.001, 0.1, 10, 100, 1000, 10000], 
            'fit_intercept':True,
            'copy_X':True,
            'max_iter':None,
            'tol':0.001,
            'solver':'auto',
            'random_state':None
            }
        

    elif isinstance_double(estimator, linear_model.RidgeCV):
        param_grid = {
            'alphas':(1.0,0.001, 0.1, 10, 100, 1000, 10000),
            'fit_intercept':True,
            'scoring':None,
            'gcv_mode':None,
            'store_cv_values':False,
            'alpha_per_target':False
            }
        
    # elif isinstance_double(estimator, linear_model.SGDRegressor):
    #     raise NotImplementedError()    
        
    elif isinstance_double(estimator, linear_model.ElasticNet):
        param_grid = {
            'alpha':np.exp(np.linspace(start=np.log(100), stop=np.log(100*0.000001), num=1000)),
            'l1_ratio':[0.5,1/4,3/4,1],
            'fit_intercept':True,
            'precompute':False,
            'max_iter':10,
            'tol':0.0001,
            'warm_start':False,
            'positive':False,
            'selection':'cyclic'
            }

    elif isinstance_double(estimator, linear_model.ElasticNetCV):
        param_grid = {
            'l1_ratio':[0.5,1/4,3/4,1],
            'eps':0.001,
            'n_alphas':1000,
            'fit_intercept':True,
            'precompute':'auto',
            'max_iter':1000,
            'tol':0.0001,
            'copy_X':True,
            'positive':False,
            'selection':'cyclic'
            }

    elif isinstance_double(estimator, linear_model.Lars):        
        param_grid = {
            'fit_intercept':True,
            'precompute':'auto',
            'n_nonzero_coefs':[500,100,200,300,400,np.inf],
            'eps':2.220446049250313e-16,
            'copy_X':True,
            'fit_path':True,
            'jitter':None,
            }

    elif isinstance_double(estimator, linear_model.LarsCV):
        param_grid = {
            'fit_intercept':True,
            'max_iter':500,
            'precompute':'auto',
            'max_n_alphas':1000,
            'eps':2.220446049250313e-16,
            'copy_X':True
            }
        
    elif isinstance_double(estimator, linear_model.Lasso):
        param_grid = {
            'alpha':np.exp(np.linspace(start=np.log(100), stop=np.log(100*0.000001), num=1000)),
            'fit_intercept':True,
            'precompute':False,
            'copy_X':True,
            'max_iter':1000,
            'tol':0.0001,
            'warm_start':False,
            'positive':False,
            'selection':'cyclic'
            }

    elif isinstance_double(estimator, linear_model.LassoCV):
        param_grid = {
            'eps':0.001,
            'n_alphas':1000,
            'fit_intercept':True,
            'precompute':'auto',
            'max_iter':1000,
            'tol':0.0001,
            'copy_X':True,
            'positive':False,
            'selection':'cyclic'
            }

    elif isinstance_double(estimator, linear_model.LassoLars):
        param_grid = {
            'alpha':np.exp(np.linspace(start=np.log(100), stop=np.log(100*0.000001), num=1000)),
            'fit_intercept':True,
            'precompute':'auto',
            'max_iter':500,
            'eps':2.220446049250313e-16,
            'copy_X':True,
            'fit_path':True,
            'positive':False,
            'jitter':None,
            }

    elif isinstance_double(estimator, linear_model.LassoLarsCV):
        param_grid = {
            'fit_intercept':True,
            'max_iter':500,
            'precompute':'auto',
            'max_n_alphas':1000,
            'eps':2.220446049250313e-16,
            'copy_X':True,
            'positive':False
            }
        
    # elif isinstance_double(estimator, linear_model.OrthogonalMatchingPursuit):
    #     raise NotImplementedError()            
        
    # elif isinstance_double(estimator, linear_model.ARDRegression):
    #     raise NotImplementedError()           

    # elif isinstance_double(estimator, linear_model.BayesianRidge):
    #     raise NotImplementedError()           

    # elif isinstance_double(estimator, linear_model.HuberRegressor):
    #     raise NotImplementedError()           

    # elif isinstance_double(estimator, linear_model.QuantileRegressor):
    #     raise NotImplementedError()           

    # elif isinstance_double(estimator, linear_model.RANSACRegressor):
    #     raise NotImplementedError()           

    # elif isinstance_double(estimator, linear_model.TheilSenRegressor):
    #     raise NotImplementedError()                   
        
    #--------------------------------------------------------------------------
    # from sklearn.neighbors
    #--------------------------------------------------------------------------
    
    #--------------------------------------------------------------------------
    # from sklearn.neural_network
    #--------------------------------------------------------------------------
    elif isinstance_double(estimator, neural_network.MLPRegressor):
        param_grid = {
            'hidden_layer_sizes':[(100,), (2,), (2,4), (2,4,8,), (2,4,8,16,), (2,4,8,16,32),
                                          (4,), (4,8), (4,8,16), (4,8,16,32),
                                          (8,), (8,16), (8,16,32),
                                          (16,), (16,32),
                                          (32,)],
            'activation':'relu',
            'solver':'adam',
            'alpha':[0.0001,1.e+01, 1.e+02, 1.e+03, 1.e+05, 1.e+06],
            'batch_size':'auto',
            'learning_rate':['constant', 'invscaling', 'adaptive'],
            'learning_rate_init':0.001,
            'power_t':0.5,
            'max_iter':200,
            'shuffle':True,
            'tol':0.0001,
            'momentum':0.9,
            'nesterovs_momentum':True,
            'early_stopping':True,
            'validation_fraction':0.1,
            'beta_1':0.9,
            'beta_2':0.999,
            'epsilon':1e-08,
            'n_iter_no_change':10,
            'max_fun':15000
            }
        
    #--------------------------------------------------------------------------
    # from sklearn.svm
    #--------------------------------------------------------------------------
    
    #--------------------------------------------------------------------------
    # from sklearn.tree
    #--------------------------------------------------------------------------

    #--------------------------------------------------------------------------
    # Other
    #--------------------------------------------------------------------------
    else:
        try:
            param_grid = estimator.get_params()
        except:        
            raise NotImplementedError(f"Algorithm '{estimator}' is currently not implemented and has no 'get_params()'-method")
            
    return param_grid

  
    #     },
    #     "SGDRegressor" : {
    #         "model_params" : {
    #             "loss":'squared_loss',
    #             "penalty":'l2',
    #             "l1_ratio":0.15,
    #             "fit_intercept":True,
    #             "max_iter":10000,
    #             "tol":0.0001,
    #             "shuffle":True,
    #             "verbose":0,
    #             "epsilon":0.1,
    #             "random_state":None,
    #             "eta0":0.01,
    #             "power_t":0.25,
    #             "early_stopping":True,
    #             "validation_fraction":0.1,
    #             "n_iter_no_change":50,
    #             "warm_start":False, 
    #             "average":False
    #             },
    #         "tuning_params" : {
    #             "alpha":[0.0001, 0.001, 0.01, 0.00001, 0.000001],
    #             "learning_rate":['invscaling', 'constant', 'optimal', 'adaptive'],
    #             }
    #     },
    #     "ARDRegression" : {
    #         "model_params" : {
    #             "n_iter":1000,
    #             "tol":0.0001,
    #             "compute_score":False,
    #             "threshold_lambda":10000.0,
    #             "fit_intercept":True,
    #             "normalize":False,
    #             "copy_X":True,
    #             "verbose":False
    #             },
    #         "tuning_params" : {
    #             "alpha_1":[1e-06, 1e-07, 1e-08, 1e-05, 1e-04],
    #             "alpha_2":[1e-06, 1e-07, 1e-08, 1e-05, 1e-04],
    #             "lambda_1":[1e-06, 1e-07, 1e-08, 1e-05, 1e-04],
    #             "lambda_2":[1e-06, 1e-07, 1e-08, 1e-05, 1e-04],
    #             }
    #     },
    #     "BayesianRidge" : {
    #         "model_params" : {
    #             "n_iter":1000,
    #             "tol":0.0001,
    #             'alpha_init':None,
    #             'lambda_init':None,
    #             "compute_score":False,
    #             "fit_intercept":True,
    #             "normalize":False,
    #             "copy_X":True,
    #             "verbose":False
    #             },
    #         "tuning_params" : {
    #             "alpha_1":[1e-06, 1e-07, 1e-08, 1e-05, 1e-04],
    #             "alpha_2":[1e-06, 1e-07, 1e-08, 1e-05, 1e-04],
    #             "lambda_1":[1e-06, 1e-07, 1e-08, 1e-05, 1e-04],
    #             "lambda_2":[1e-06, 1e-07, 1e-08, 1e-05, 1e-04],
    #             }
    #     },
    #     "HuberRegressor" : {
    #         "model_params" : {
    #             "max_iter":10000,
    #             "warm_start":False,
    #             "fit_intercept":True,
    #             "tol":0.00001
    #             },
    #         "tuning_params" : {
    #             "epsilon":[1.35, 1.1,1.2,1.3,1.4,1.5,175],
    #             "alpha":[0.0001, 0.001, 0.01, 0.00001, 0.000001],                                        
    #             }
    #     },
    #     "RANSACRegressor" : {
    #         "model_params" : {
    #             "min_samples":None,
    #             "residual_threshold":None, 
    #             "is_data_valid":None,
    #             "is_model_valid":None,
    #             "max_trials":100,
    #             "max_skips":np.inf,
    #             "stop_n_inliers":np.inf,
    #             "stop_score":np.inf,
    #             "stop_probability":0.99,
    #             "loss":'absolute_loss',
    #             "random_state":None
    #             },
    #         "tuning_params" : {
    #             "base_estimator":linear_model.LinearRegression(),
    #             }
    #     },
    #     "TheilSenRegressor" : {
    #         "model_params" : {
    #             "fit_intercept":True,
    #             "copy_X":True,
    #             "max_subpopulation":10000,
    #             "n_subsamples":None,
    #             "max_iter":10000,
    #             "tol":0.0001,
    #             "random_state":None,
    #             "n_jobs":None,
    #             "verbose":False
    #             },
    #         "tuning_params" : {}
    #     },
    #     "PassiveAggressiveRegressor" : {
    #         "model_params" : {
    #             "fit_intercept":True,
    #             "max_iter":10000,
    #             "tol":0.0001,
    #             "early_stopping":True,
    #             "validation_fraction":0.1,
    #             "n_iter_no_change":50,
    #             "shuffle":True,
    #             "verbose":0,
    #             "loss":'epsilon_insensitive',
    #             "epsilon":0.1,
    #             "random_state":None,
    #             "warm_start":False,
    #             "average":False
    #             },
    #         "tuning_params" : {
    #             "C":[0.1,1.0,10],
    #             }
    #     },
    #     "KNeighborsRegressor" : {
    #         "model_params" : {
    #             "algorithm":'auto',
    #             "leaf_size":30,
    #             "p":2,
    #             "metric":'minkowski',
    #             "metric_params":None,
    #             "n_jobs":None
    #             },
    #         "tuning_params" : {
    #             "n_neighbors":[5,1,2,4,8,16,32],
    #             "weights":['distance','uniform'],
    #             }
    #
    #     },
    #     "LinearSVR" : {
    #         "model_params" : {
    #             "epsilon":0.0,
    #             "tol":0.0001,
    #             "loss":'epsilon_insensitive',
    #             "fit_intercept":True,
    #             "intercept_scaling":1.0,
    #             "dual":dual_par,
    #             "verbose":0,
    #             "random_state":None,
    #             "max_iter":1000000
    #             },
    #         "tuning_params" : {
    #             "C":[0.1,1,2,5,10],
    #             }
    #     },
    #     "NuSVR" : {
    #         "model_params" : {
    #             "degree":3,
    #             "gamma":'scale',
    #             "coef0":0.0,
    #             "shrinking":True,
    #             "tol":0.001,
    #             "cache_size":200,
    #             "verbose":False,
    #             "max_iter":- 1
    #             },
    #         "tuning_params" : {
    #             "C":[0.1,1,2,5,10],
    #             "nu":[1/4,1/2, 3/4],
    #             "kernel":['linear', 'poly', 'rbf', 'sigmoid'],
    #             }
    #     }                          
    # }
